/*
 GBxCart RW - Console Interface
 Version: 1.20
 Author: Alex from insideGadgets (www.insidegadgets.com)
 Created: 7/11/2016
 Last Modified: 26/01/2019
 
 GBxCart RW allows you to dump your Gameboy/Gameboy Colour/Gameboy Advance games ROM, save the RAM and write to the RAM.
 
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <Windows.h>
#else
#include <unistd.h>
#endif

#include "setup.h" // See defines, variables, constants, functions here

int main(int argc, char **argv) {
	
	printf("GBxCart RW v1.20 by insideGadgets\n");
	printf("################################\n");
	
	read_config();
	
	// Open COM port
	if (com_test_port() == 0) {
		printf("Device not connected and couldn't be auto detected\n");
		read_one_letter();
		return 1;
	}
	printf("Connected on COM port: %i\n", cport_nr+1);
	
	// Break out of any existing functions on ATmega
	set_mode('0');
	
	// Get firmware version
	gbxcartFirmwareVersion = request_value(READ_FIRMWARE_VERSION);
	
	// Get PCB version
	gbxcartPcbVersion = request_value(READ_PCB_VERSION);
	
	// Prompt for GB or GBA mode
	if (gbxcartPcbVersion == PCB_1_3) {
		set_mode(VOLTAGE_5V);
		delay_ms(300);
	}
	
	
	gb_flash_pin_setup(WE_AS_WR_PIN); // WR pin
	read_gb_header();
	cartridgeType = 8;
	
	gb_flash_write_address_byte(0x7000, 0x00);
	gb_flash_write_address_byte(0x7001, 0x00);
	gb_flash_write_address_byte(0x7002, 0x93);
	delay_ms(1);
	
	romBanks = 512;
	strncpy(gameTitle, "BANK3", 6);
	
	
	// Dump ROM		
	printf("\n--- Read ROM ---\n");
	
	char titleFilename[30];
	strncpy(titleFilename, gameTitle, 20);
	if (cartridgeMode == GB_MODE) {
		strncat(titleFilename, ".gb", 3);
	}
	else {
		strncat(titleFilename, ".gba", 4);
	}
	printf("Reading ROM to %s\n", titleFilename);
	printf("[             25%%             50%%             75%%            100%%]\n[");
	
	// Create a new file
	FILE *romFile = fopen(titleFilename, "wb");
	
	uint32_t readBytes = 0;
	if (cartridgeMode == GB_MODE) {
		// Set start and end address
		currAddr = 0x0000;
		endAddr = 0x7FFF;
		
		// Read ROM
		for (uint16_t bank = 1; bank < romBanks; bank++) {				
			if (cartridgeType >= 5) { // MBC2 and above
				set_bank(0x2100, bank);
				if (bank >= 256) {
					set_bank(0x3000, 1); // High bit
				}
				else {
					set_bank(0x3000, 0); // High bit
				}
			}
			else { // MBC1
				set_bank(0x6000, 0); // Set ROM Mode 
				set_bank(0x4000, bank >> 5); // Set bits 5 & 6 (01100000) of ROM bank
				set_bank(0x2000, bank & 0x1F); // Set bits 0 & 4 (00011111) of ROM bank
			}
			
			if (bank > 1) { currAddr = 0x4000; }
			
			// Set start address and rom reading mode
			set_number(currAddr, SET_START_ADDRESS);
			set_mode(READ_ROM_RAM);
			
			// Read data
			while (currAddr < endAddr) {
				com_read_bytes(romFile, 64);
				currAddr += 64;
				readBytes += 64;
				
				// Request 64 bytes more
				if (currAddr < endAddr) {
					com_read_cont();
				}
				
				// Print progress
				print_progress_percent(readBytes, (romBanks * 16384) / 64);
			}
			com_read_stop(); // Stop reading ROM (as we will bank switch)
		}
		printf("]");
	}
	
	fclose(romFile);
	printf("\nFinished\n");
	
	
	
	
	return 0;
}